#!/usr/bin/env node
var path = require('path');
var arg = require('arg');
var fs = require('fs/promises');

function _interopDefault (e) { return e && e.__esModule ? e : { default: e }; }

var path__default = /*#__PURE__*/_interopDefault(path);
var arg__default = /*#__PURE__*/_interopDefault(arg);
var fs__default = /*#__PURE__*/_interopDefault(fs);

function exit(err) {
    logger.error(err);
    process.exit(1);
}
const formatDuration = (duration)=>duration >= 1000 ? `${duration / 1000}s` : `${duration}ms`;
async function hasPackageJson(cwd) {
    return await fileExists(path__default.default.resolve(cwd, 'package.json'));
}
async function getPackageMeta(cwd) {
    const pkgFilePath = path__default.default.resolve(cwd, 'package.json');
    let targetPackageJson = {};
    try {
        targetPackageJson = JSON.parse(await fs__default.default.readFile(pkgFilePath, {
            encoding: 'utf-8'
        }));
    } catch (_) {}
    return targetPackageJson;
}
const logger = {
    log (arg) {
        console.log(arg);
    },
    warn (arg) {
        console.log('\x1b[33m' + arg + '\x1b[0m');
    },
    error (arg) {
        console.error('\x1b[31m' + (arg instanceof Error ? arg.stack : arg) + '\x1b[0m');
    }
};
async function fileExists(filePath) {
    try {
        await fs__default.default.access(filePath);
        return true;
    } catch (err) {
        if (err.code === 'ENOENT') {
            return false;
        }
        throw err;
    }
}

var version = "3.7.0";

const helpMessage = `
Usage: bunchee [options]

Options:
  -v, --version          output the version number
  -w, --watch            watch src files changes
  -m, --minify           compress output. default: false
  -o, --output <file>    specify output filename
  -f, --format <format>  type of output (esm, amd, cjs, iife, umd, system), default: esm
  -h, --help             output usage information
  --external <mod>       specify an external dependency, separate by comma
  --no-external          do not bundle external dependencies
  --target <target>      js features target: swc target es versions. default: es2016
  --runtime <runtime>    build runtime (nodejs, browser). default: browser
  --env <env>            inlined process env variables, separate by comma. default: NODE_ENV
  --cwd <cwd>            specify current working directory
  --sourcemap            enable sourcemap generation, default: false
  --dts                  determine if need to generate types, default: false
`;
function help() {
    logger.log(helpMessage);
}
async function lintPackage(cwd) {
    // Not package.json detected, skip package linting
    if (!await hasPackageJson(cwd)) {
        return;
    }
    const { publint } = await import('publint');
    const { printMessage } = await import('publint/utils');
    const messages = await publint({
        pkgDir: cwd,
        level: 'error'
    });
    const pkg = await getPackageMeta(cwd);
    for (const message of messages){
        console.log(printMessage(message, pkg));
    }
}
function parseCliArgs(argv) {
    let args;
    args = arg__default.default({
        '--cwd': String,
        '--dts': Boolean,
        '--output': String,
        '--format': String,
        '--watch': Boolean,
        '--minify': Boolean,
        '--help': Boolean,
        '--version': Boolean,
        '--runtime': String,
        '--target': String,
        '--sourcemap': Boolean,
        '--env': String,
        '--external': String,
        '--no-external': Boolean,
        '-h': '--help',
        '-v': '--version',
        '-w': '--watch',
        '-o': '--output',
        '-f': '--format',
        '-m': '--minify'
    }, {
        permissive: true,
        argv
    });
    const source = args._[0];
    const parsedArgs = {
        source,
        format: args['--format'],
        file: args['--output'],
        watch: args['--watch'],
        minify: args['--minify'],
        sourcemap: !!args['--sourcemap'],
        cwd: args['--cwd'],
        dts: args['--dts'],
        help: args['--help'],
        version: args['--version'],
        runtime: args['--runtime'],
        target: args['--target'],
        external: !!args['--no-external'] ? null : args['--external'],
        env: args['--env']
    };
    return parsedArgs;
}
async function run(args) {
    var _args_external, _env;
    const { source, format, watch, minify, sourcemap, target, runtime, dts, env } = args;
    const cwd = args.cwd || process.cwd();
    const file = args.file ? path__default.default.resolve(cwd, args.file) : undefined;
    const bundleConfig = {
        dts,
        file,
        format,
        cwd,
        target,
        runtime,
        external: ((_args_external = args.external) == null ? void 0 : _args_external.split(',')) || [],
        watch: !!watch,
        minify: !!minify,
        sourcemap: sourcemap === false ? false : true,
        env: ((_env = env) == null ? void 0 : _env.split(',')) || []
    };
    if (args.version) {
        return logger.log(version);
    }
    if (args.help) {
        return help();
    }
    const entry = source ? path__default.default.resolve(cwd, source) : '';
    const bundle = require('./index').bundle;
    let timeStart = Date.now();
    let timeEnd;
    try {
        await bundle(entry, bundleConfig);
        timeEnd = Date.now();
    } catch (err) {
        if (err.name === 'NOT_EXISTED') {
            help();
            return exit(err);
        }
        throw err;
    }
    const duration = timeEnd - timeStart;
    // watching mode
    if (watch) {
        logger.log(`🔍  Watching assets in ${cwd}...`);
        return;
    }
    // build mode
    logger.log(`✨  Finished in ${formatDuration(duration)}`);
    await lintPackage(cwd);
}
async function main() {
    let params, error;
    try {
        params = parseCliArgs(process.argv.slice(2));
    } catch (err) {
        error = err;
    }
    if (error || !params) {
        if (!error) help();
        return exit(error);
    }
    await run(params);
}
main().catch(exit);
