Object.defineProperty(exports, '__esModule', { value: true });

var fs = require('fs/promises');
var path = require('path');
var rollup = require('rollup');
var pluginWasm = require('@rollup/plugin-wasm');
var rollupPluginSwc3 = require('rollup-plugin-swc3');
var commonjs = require('@rollup/plugin-commonjs');
var json = require('@rollup/plugin-json');
var pluginNodeResolve = require('@rollup/plugin-node-resolve');
var replace = require('@rollup/plugin-replace');
var prettyBytes = require('pretty-bytes');
var swcPreserveDirectivePlugin = require('rollup-swc-preserve-directives');
var module$1 = require('module');

function _interopDefault (e) { return e && e.__esModule ? e : { default: e }; }

var fs__default = /*#__PURE__*/_interopDefault(fs);
var path__default = /*#__PURE__*/_interopDefault(path);
var commonjs__default = /*#__PURE__*/_interopDefault(commonjs);
var json__default = /*#__PURE__*/_interopDefault(json);
var replace__default = /*#__PURE__*/_interopDefault(replace);
var prettyBytes__default = /*#__PURE__*/_interopDefault(prettyBytes);
var swcPreserveDirectivePlugin__default = /*#__PURE__*/_interopDefault(swcPreserveDirectivePlugin);

function chunkSizeCollector() {
    const sizes = new Map();
    function addSize(name, size) {
        sizes.set(name, size);
    }
    return {
        plugin: (cwd)=>{
            return {
                name: 'collect-sizes',
                augmentChunkHash () {
                // Do nothing, but use the hook to keep the plugin instance alive
                },
                renderChunk (code, chunk, options) {
                    const dir = options.dir || options.file && path__default.default.dirname(options.file);
                    let fileName = chunk.fileName;
                    if (dir) {
                        const filePath = path__default.default.join(dir, fileName);
                        fileName = filePath.startsWith(cwd) ? path__default.default.relative(cwd, filePath) : filePath;
                    }
                    addSize(fileName, code.length);
                    return null;
                }
            };
        },
        getSizeStats () {
            const sizeStats = [];
            sizes.forEach((size, name)=>{
                sizeStats.push([
                    name,
                    prettyBytes__default.default(size),
                    size
                ]);
            });
            return sizeStats;
        }
    };
}

const availableExtensions = [
    'js',
    'cjs',
    'mjs',
    'jsx',
    'ts',
    'tsx',
    'cts',
    'mts'
];
const availableExportConventions = [
    'react-server',
    'react-native',
    'edge-light'
];
const availableESExtensionsRegex = /\.(m|c)?[jt]sx?$/;
const dtsExtensionRegex = /\.d\.(m|c)?ts$/;
const SRC = 'src';

function exit(err) {
    logger.error(err);
    process.exit(1);
}
async function getPackageMeta(cwd) {
    const pkgFilePath = path__default.default.resolve(cwd, 'package.json');
    let targetPackageJson = {};
    try {
        targetPackageJson = JSON.parse(await fs__default.default.readFile(pkgFilePath, {
            encoding: 'utf-8'
        }));
    } catch (_) {}
    return targetPackageJson;
}
const logger = {
    log (arg) {
        console.log(arg);
    },
    warn (arg) {
        console.log('\x1b[33m' + arg + '\x1b[0m');
    },
    error (arg) {
        console.error('\x1b[31m' + (arg instanceof Error ? arg.stack : arg) + '\x1b[0m');
    }
};
async function fileExists(filePath) {
    try {
        await fs__default.default.access(filePath);
        return true;
    } catch (err) {
        if (err.code === 'ENOENT') {
            return false;
        }
        throw err;
    }
}
// . -> pkg name
// ./lite -> <pkg name>/lite
function getExportPath(pkg, cwd, exportName) {
    const name = pkg.name || path__default.default.basename(cwd);
    if (exportName === '.' || !exportName) return name;
    return path__default.default.join(name, exportName);
}
const isNotNull = (n)=>Boolean(n);
function resolveSourceFile(cwd, filename) {
    return path__default.default.resolve(cwd, SRC, filename);
}
async function findSourceEntryFile(cwd, exportPath, exportTypeSuffix, ext) {
    const filename = resolveSourceFile(cwd, `${exportPath}${exportTypeSuffix ? `.${exportTypeSuffix}` : ''}.${ext}`);
    if (await fileExists(filename)) {
        return filename;
    }
    const subFolderIndexFilename = resolveSourceFile(cwd, `${exportPath}/index${exportTypeSuffix ? `.${exportTypeSuffix}` : ''}.${ext}`);
    if (await fileExists(subFolderIndexFilename)) {
        return subFolderIndexFilename;
    }
    return undefined;
}
// Map '.' -> './index.[ext]'
// Map './lite' -> './lite.[ext]'
// Return undefined if no match or if it's package.json exports
async function getSourcePathFromExportPath(cwd, exportPath, exportType) {
    for (const ext of availableExtensions){
        // ignore package.json
        if (exportPath.endsWith('package.json')) return;
        if (exportPath === '.') exportPath = './index';
        // Find convention-based source file for specific export types
        if (availableExportConventions.includes(exportType)) {
            const filename = await findSourceEntryFile(cwd, exportPath, exportType, ext);
            if (filename) return filename;
        }
        const filename = await findSourceEntryFile(cwd, exportPath, null, ext);
        if (filename) return filename;
    }
    return;
}
// Unlike path.basename, forcedly removing extension
function filenameWithoutExtension(file) {
    return file ? file.replace(new RegExp(`${path__default.default.extname(file)}$`), '') : undefined;
}
const nonNullable = (n)=>Boolean(n);
const hasAvailableExtension = (filename)=>availableExtensions.includes(path__default.default.extname(filename).slice(1));

function getTypings(pkg) {
    return pkg.types || pkg.typings;
}
function getDistPath(distPath, cwd) {
    return path.resolve(cwd, distPath);
}
// Reached the end of the export path
function isExportLike(field) {
    if (typeof field === 'string') return true;
    return Object.entries(field).every(// Every value is string and key is not start with '.'
    // TODO: check key is ExportType
    ([key, value])=>typeof value === 'string' && !key.startsWith('.'));
}
function constructFullExportCondition(exportCondition, packageType) {
    const isCommonjs = packageType === 'commonjs';
    let result;
    if (typeof exportCondition === 'string') {
        result = {
            [isCommonjs ? 'require' : 'import']: exportCondition
        };
    } else {
        // TODO: valid export condition, warn if it's not valid
        const keys = Object.keys(exportCondition);
        result = {};
        keys.forEach((key)=>{
            const condition = exportCondition[key];
            // Filter out nullable value
            if (key in exportCondition && condition) {
                result[key] = condition;
            }
        });
    }
    return result;
}
function joinRelativePath(...segments) {
    let result = path.join(...segments);
    // If the first segment starts with './', ensure the result does too.
    if (segments[0] === '.' && !result.startsWith('./')) {
        result = './' + result;
    }
    return result;
}
function findExport(name, exportCondition, paths, packageType) {
    // TODO: handle export condition based on package.type
    if (isExportLike(exportCondition)) {
        paths[name] = constructFullExportCondition(exportCondition, packageType);
        return;
    }
    Object.keys(exportCondition).forEach((subpath)=>{
        const nextName = joinRelativePath(name, subpath);
        const nestedExportCondition = exportCondition[subpath];
        findExport(nextName, nestedExportCondition, paths, packageType);
    });
}
/**
 *
 * Convert package.exports field to paths mapping
 * Example
 *
 * Input:
 * {
 *   ".": {
 *     "sub": {
 *       "import": "./sub.js",
 *       "require": "./sub.cjs",
 *       "types": "./sub.d.ts
 *     }
 *   }
 * }
 *
 * Output:
 * {
 *   "./sub": {
 *     "import": "./sub.js",
 *     "require": "./sub.cjs",
 *     "types": "./sub.d.ts,
 *   }
 * }
 *
 */ function parseExport(exportsCondition, packageType) {
    const paths = {};
    if (typeof exportsCondition === 'string') {
        paths['.'] = constructFullExportCondition(exportsCondition, packageType);
    } else if (typeof exportsCondition === 'object') {
        if (isExportLike(exportsCondition)) {
            paths['.'] = constructFullExportCondition(exportsCondition, packageType);
        } else {
            Object.keys(exportsCondition).forEach((key)=>{
                const exportCondition = exportsCondition[key];
                findExport(key, exportCondition, paths, packageType);
            });
        }
    }
    return paths;
}
/**
 * Get package exports paths
 *
 * Example:
 *
 * ```json
 * {
 *  "exports": {
 *    ".": {
 *      "require": "./dist/index.cjs",
 *      "module": "./dist/index.esm.js",
 *      "default": "./dist/index.esm.js"
 *    },
 *    "./foo": {
 *      "require": "./dist/foo.cjs",
 *      "module": "./dist/foo.esm.js",
 *      "default": "./dist/foo.esm.js"
 *   }
 * }
 *
 * ```
 *
 * will be parsed to:
 *
 * ```js
 * {
 *   '.': {
 *     main: './dist/index.cjs',
 *     module: './dist/index.esm.js',
 *     export: './dist/index.esm.js'
 *   },
 *   './foo': {
 *     main: './dist/foo.cjs',
 *     module: './dist/foo.esm.js',
 *     export: './dist/foo.esm.js'
 *   }
 *
 *
 * pkg.main and pkg.module will be added to ['.'] if exists
 */ function getExportPaths(pkg, pkgType, resolvedWildcardExports) {
    var _pathsMap_;
    const pathsMap = {};
    const packageType = pkgType != null ? pkgType : getPackageType(pkg);
    const isCjsPackage = packageType === 'commonjs';
    const exportsConditions = resolvedWildcardExports != null ? resolvedWildcardExports : pkg.exports;
    if (exportsConditions) {
        const paths = parseExport(exportsConditions, packageType);
        Object.assign(pathsMap, paths);
    }
    // main export '.' from main/module/typings
    const defaultMainExport = constructFullExportCondition({
        [isCjsPackage ? 'require' : 'import']: pkg.main,
        module: pkg.module,
        types: getTypings(pkg)
    }, packageType);
    if (isCjsPackage && ((_pathsMap_ = pathsMap['.']) == null ? void 0 : _pathsMap_['require'])) {
        // pathsMap's exports.require are prioritized.
        defaultMainExport['require'] = pathsMap['.']['require'];
    }
    // Merge the main export into '.' paths
    const mainExport = Object.assign({}, pathsMap['.'], defaultMainExport);
    // main export is not empty
    if (Object.keys(mainExport).length > 0) {
        pathsMap['.'] = mainExport;
    }
    return pathsMap;
}
const getExportTypeDist = (parsedExportCondition, cwd)=>{
    const existed = new Set();
    const exportTypes = Object.keys(parsedExportCondition.export);
    for (const key of exportTypes){
        if (key === 'module') {
            continue;
        }
        const filePath = parsedExportCondition.export[key];
        if (key === 'types') {
            const typeFile = getDistPath(filePath, cwd);
            if (existed.has(typeFile)) {
                continue;
            }
            existed.add(typeFile);
            continue;
        }
        const ext = path.extname(filePath).slice(1);
        const dtsExtentions = {
            js: '.d.ts',
            cjs: '.d.cts',
            mjs: '.d.mts'
        };
        const typeFile = getDistPath(`${filenameWithoutExtension(filePath) || ''}${dtsExtentions[ext]}`, cwd);
        if (existed.has(typeFile)) {
            continue;
        }
        existed.add(typeFile);
    }
    return Array.from(existed);
};
function getPackageType(pkg) {
    return pkg.type || 'commonjs';
}
function constructDefaultExportCondition(value, packageType) {
    let exportCondition;
    if (typeof value === 'string') {
        const types = getTypings(value);
        exportCondition = {
            [packageType === 'commonjs' ? 'require' : 'import']: value,
            ...types && {
                types
            }
        };
    } else {
        exportCondition = value;
    }
    return constructFullExportCondition(exportCondition, packageType);
}
function isEsmExportName(name, ext) {
    return [
        'import',
        'module'
    ].includes(name) || ext === 'mjs';
}
function isCjsExportName(name, ext) {
    return [
        'require',
        'main',
        'node',
        'default'
    ].includes(name) || ext === 'cjs';
}
function getExportConditionDist(pkg, parsedExportCondition, cwd) {
    const dist = [];
    const existed = new Set();
    const exportTypes = Object.keys(parsedExportCondition.export);
    for (const key of exportTypes){
        if (key === 'types') {
            continue;
        }
        const filePath = parsedExportCondition.export[key];
        const ext = path.extname(filePath).slice(1);
        const relativePath = parsedExportCondition.export[key];
        const distFile = getDistPath(relativePath, cwd);
        let format = 'esm';
        if (isEsmExportName(key, ext)) {
            format = 'esm';
        } else if (isCjsExportName(key, ext)) {
            format = 'cjs';
        }
        // Deduplicate the same path jobs
        // TODO: detect conflicts paths but with different format
        if (existed.has(distFile)) {
            continue;
        }
        existed.add(distFile);
        dist.push({
            format,
            file: distFile
        });
    }
    if (dist.length === 0) {
        // TODO: Deprecate this warning and behavior in v3
        console.error(`Doesn't fin any exports in ${pkg.name}, using default dist path dist/index.js`);
        dist.push({
            format: 'esm',
            file: getDistPath('dist/index.js', cwd)
        });
    }
    return dist;
}
function getTypeFilePath(entryFilePath, exportCondition, cwd) {
    var _exportCondition;
    const name = filenameWithoutExtension(entryFilePath);
    const firstDistPath = exportCondition ? Object.values(exportCondition.export)[0] : undefined;
    const exportName = ((_exportCondition = exportCondition) == null ? void 0 : _exportCondition.name) || 'index';
    return entryFilePath ? name + '.d.ts' : path.resolve(firstDistPath ? path.dirname(firstDistPath) : path.join(cwd, 'dist'), (exportName === '.' ? 'index' : exportName) + '.d.ts');
}

const swcMinifyOptions = {
    compress: true,
    format: {
        comments: 'some',
        wrapFuncArgs: false,
        preserveAnnotations: true
    },
    mangle: {
        toplevel: true
    }
};
// This can also be passed down as stats from top level
const sizeCollector = chunkSizeCollector();
function getBuildEnv(envs) {
    if (!envs.includes('NODE_ENV')) {
        envs.push('NODE_ENV');
    }
    const envVars = envs.reduce((acc, key)=>{
        const value = process.env[key];
        if (typeof value !== 'undefined') {
            acc['process.env.' + key] = JSON.stringify(value);
        }
        return acc;
    }, {});
    return envVars;
}
function buildInputConfig(entry, pkg, options, cwd, { tsConfigPath, tsCompilerOptions }, dts) {
    var _tsCompilerOptions;
    var _options_external;
    const externals = options.noExternal ? [] : [
        pkg.peerDependencies,
        pkg.dependencies,
        pkg.peerDependenciesMeta
    ].filter((n)=>Boolean(n)).map((o)=>Object.keys(o)).reduce((a, b)=>a.concat(b), []).concat(((_options_external = options.external) != null ? _options_external : []).concat(pkg.name ? [
        pkg.name
    ] : []));
    const { useTypescript, runtime, target: jscTarget, minify: shouldMinify } = options;
    const hasSpecifiedTsTarget = Boolean(((_tsCompilerOptions = tsCompilerOptions) == null ? void 0 : _tsCompilerOptions.target) && tsConfigPath);
    const swcParserConfig = {
        syntax: useTypescript ? 'typescript' : 'ecmascript',
        [useTypescript ? 'tsx' : 'jsx']: true,
        privateMethod: true,
        classPrivateProperty: true,
        exportDefaultFrom: true
    };
    const swcOptions = {
        jsc: {
            ...!hasSpecifiedTsTarget && {
                target: jscTarget
            },
            loose: true,
            externalHelpers: false,
            parser: swcParserConfig,
            ...shouldMinify && {
                minify: {
                    ...swcMinifyOptions,
                    sourceMap: options.sourcemap
                }
            }
        },
        sourceMaps: options.sourcemap,
        inlineSourcesContent: false
    };
    const sizePlugin = sizeCollector.plugin(cwd);
    const commonPlugins = [
        swcPreserveDirectivePlugin__default.default(),
        sizePlugin
    ];
    const plugins = (dts ? [
        ...commonPlugins,
        useTypescript && require('rollup-plugin-dts').default({
            compilerOptions: {
                ...tsCompilerOptions,
                declaration: true,
                noEmit: false,
                noEmitOnError: true,
                emitDeclarationOnly: true,
                checkJs: false,
                declarationMap: false,
                skipLibCheck: true,
                preserveSymlinks: false,
                target: 'esnext',
                module: 'esnext',
                incremental: false,
                jsx: tsCompilerOptions.jsx || 'react'
            }
        })
    ] : [
        ...commonPlugins,
        replace__default.default({
            values: getBuildEnv(options.env || []),
            preventAssignment: true
        }),
        pluginNodeResolve.nodeResolve({
            preferBuiltins: runtime === 'node',
            extensions: [
                '.mjs',
                '.cjs',
                '.js',
                '.json',
                '.node',
                '.jsx'
            ]
        }),
        commonjs__default.default({
            include: /node_modules\//
        }),
        json__default.default(),
        pluginWasm.wasm(),
        rollupPluginSwc3.swc({
            include: availableESExtensionsRegex,
            exclude: 'node_modules',
            tsconfig: tsConfigPath,
            ...swcOptions
        })
    ]).filter(isNotNull);
    return {
        input: entry,
        external (id) {
            return externals.some((name)=>id === name || id.startsWith(name + '/'));
        },
        plugins,
        treeshake: {
            propertyReadSideEffects: false
        },
        onwarn (warning, warn) {
            const code = warning.code || '';
            // Some may not have types, like CLI binary
            if (dts && code === 'EMPTY_BUNDLE') return;
            if ([
                'MIXED_EXPORTS',
                'PREFER_NAMED_EXPORTS',
                'UNRESOLVED_IMPORT',
                'THIS_IS_UNDEFINED'
            ].includes(code)) return;
            // If the circular dependency warning is from node_modules, ignore it
            if (code === 'CIRCULAR_DEPENDENCY' && /Circular dependency: node_modules/.test(warning.message)) {
                return;
            }
            warn(warning);
        }
    };
}
function hasEsmExport(exportPaths, tsCompilerOptions) {
    var _tsCompilerOptions;
    let hasEsm = false;
    for(const key in exportPaths){
        const exportInfo = exportPaths[key];
        const exportInfoEntries = Object.entries(exportInfo);
        if (exportInfoEntries.some(([exportType, file])=>isEsmExportName(exportType, file))) {
            hasEsm = true;
            break;
        }
    }
    return Boolean(hasEsm || ((_tsCompilerOptions = tsCompilerOptions) == null ? void 0 : _tsCompilerOptions.esModuleInterop));
}
function buildOutputConfigs(pkg, exportPaths, options, exportCondition, cwd, { tsCompilerOptions }, dts) {
    const { format } = options;
    // Add esm mark and interop helper if esm export is detected
    const useEsModuleMark = hasEsmExport(exportPaths, tsCompilerOptions);
    const typings = getTypings(pkg);
    const file = options.file && path.resolve(cwd, options.file);
    const dtsDir = typings ? path.dirname(path.resolve(cwd, typings)) : path.resolve(cwd, 'dist');
    const name = filenameWithoutExtension(file);
    // TODO: simplify dts file name detection
    const dtsFile = file ? file : exportCondition.export['types'] ? path.resolve(cwd, exportCondition.export['types']) : path.resolve(dtsDir, (exportCondition.name === '.' ? 'index' : exportCondition.name) + '.d.ts');
    // If there's dts file, use `output.file`
    const dtsPathConfig = dtsFile ? {
        file: dtsFile
    } : {
        dir: dtsDir
    };
    return {
        name: pkg.name || name,
        ...dts ? dtsPathConfig : {
            file: file
        },
        format,
        exports: 'named',
        esModule: useEsModuleMark || 'if-default-prop',
        interop: 'auto',
        freeze: false,
        strict: false,
        sourcemap: options.sourcemap
    };
}
// build configs for each entry from package exports
async function buildEntryConfig(pkg, entryPath, exportPaths, bundleConfig, cwd, tsOptions, dts) {
    const configs = [];
    Object.keys(exportPaths).forEach(async (entryExport)=>{
        // TODO: improve the source detection
        const exportCond = exportPaths[entryExport];
        const buildConfigs = [
            createBuildConfig('', exportCond)
        ];
        // For dts job, only build the default config.
        // For assets job, build all configs.
        if (!dts) {
            if (exportCond['edge-light']) {
                buildConfigs.push(createBuildConfig('edge-light', exportCond));
            }
            if (exportCond['react-server']) {
                buildConfigs.push(createBuildConfig('react-server', exportCond));
            }
            if (exportCond['react-native']) {
                buildConfigs.push(createBuildConfig('react-native', exportCond));
            }
        }
        async function createBuildConfig(exportType, exportCondRef) {
            let exportCondForType = {
                ...exportCondRef
            };
            // Special cases of export type, only pass down the exportPaths for the type
            if (availableExportConventions.includes(exportType)) {
                exportCondForType = {
                    [entryExport]: exportCondRef[exportType]
                };
            // Basic export type, pass down the exportPaths with erasing the special ones
            } else {
                for (const exportType of availableExportConventions){
                    delete exportCondForType[exportType];
                }
            }
            const source = entryPath || await getSourcePathFromExportPath(cwd, entryExport, exportType);
            if (!source) return undefined;
            const exportCondition = {
                source,
                name: entryExport,
                export: exportCondForType
            };
            const entry = resolveSourceFile(cwd, source);
            const rollupConfig = buildConfig(entry, pkg, exportPaths, bundleConfig, exportCondition, cwd, tsOptions, dts);
            return rollupConfig;
        }
        configs.push(...buildConfigs);
    });
    return (await Promise.all(configs)).filter(nonNullable);
}
function buildConfig(entry, pkg, exportPaths, bundleConfig, exportCondition, cwd, tsOptions, dts) {
    const { file } = bundleConfig;
    const useTypescript = Boolean(tsOptions.tsConfigPath);
    const options = {
        ...bundleConfig,
        useTypescript
    };
    const inputOptions = buildInputConfig(entry, pkg, options, cwd, tsOptions, dts);
    const outputExports = getExportConditionDist(pkg, exportCondition, cwd);
    let outputConfigs = [];
    // Generate dts job - single config
    if (dts) {
        const typeOutputExports = getExportTypeDist(exportCondition, cwd);
        outputConfigs = typeOutputExports.map((v)=>buildOutputConfigs(pkg, exportPaths, {
                ...bundleConfig,
                format: 'es',
                useTypescript,
                file: v
            }, exportCondition, cwd, tsOptions, dts));
    } else {
        // multi outputs with specified format
        outputConfigs = outputExports.map((exportDist)=>{
            return buildOutputConfigs(pkg, exportPaths, {
                ...bundleConfig,
                file: exportDist.file,
                format: exportDist.format,
                useTypescript
            }, exportCondition, cwd, tsOptions, dts);
        });
        // CLI output option is always prioritized
        if (file) {
            var _outputExports_;
            const fallbackFormat = (_outputExports_ = outputExports[0]) == null ? void 0 : _outputExports_.format;
            outputConfigs = [
                buildOutputConfigs(pkg, exportPaths, {
                    ...bundleConfig,
                    file,
                    format: bundleConfig.format || fallbackFormat,
                    useTypescript
                }, exportCondition, cwd, tsOptions, dts)
            ];
        }
    }
    return {
        input: inputOptions,
        output: outputConfigs,
        exportName: exportCondition.name || '.'
    };
}

let hasLoggedTsWarning = false;
function resolveTypescript(cwd) {
    let ts;
    const m = new module$1.Module('', undefined);
    m.paths = module$1.Module._nodeModulePaths(cwd);
    try {
        ts = m.require('typescript');
    } catch (_) {
        console.error(_);
        if (!hasLoggedTsWarning) {
            hasLoggedTsWarning = true;
            exit('Could not load TypeScript compiler. Try to install `typescript` as dev dependency');
        }
    }
    return ts;
}
async function resolveTsConfig(cwd) {
    let tsCompilerOptions = {};
    let tsConfigPath;
    tsConfigPath = path.resolve(cwd, 'tsconfig.json');
    if (await fileExists(tsConfigPath)) {
        const ts = resolveTypescript(cwd);
        const basePath = tsConfigPath ? path.dirname(tsConfigPath) : cwd;
        const tsconfigJSON = ts.readConfigFile(tsConfigPath, ts.sys.readFile).config;
        tsCompilerOptions = ts.parseJsonConfigFileContent(tsconfigJSON, ts.sys, basePath).options;
    } else {
        tsConfigPath = undefined;
        return null;
    }
    return {
        tsCompilerOptions,
        tsConfigPath
    };
}

function logSizeStats() {
    const stats = sizeCollector.getSizeStats();
    const maxLength = Math.max(...stats.map(([filename])=>filename.length));
    stats.forEach(([filename, prettiedSize])=>{
        const padding = ' '.repeat(maxLength - filename.length);
        const action = dtsExtensionRegex.test(filename) ? 'Typed' : 'Built';
        logger.log(` ✓  ${action} ${filename}${padding} - ${prettiedSize}`);
    });
}

// TODO: support nested wildcard exportsCondition (e.g. './foo/*')
const getWildcardExports = (exportsCondition)=>{
    return {
        './*': exportsCondition['./*']
    };
};
const isExportable = async (dirent, pathname)=>{
    if (dirent.isDirectory()) {
        const innerDirents = await fs__default.default.readdir(path__default.default.join(pathname, dirent.name), {
            withFileTypes: true
        });
        return innerDirents.some(({ name })=>name.startsWith('index') && hasAvailableExtension(name));
    }
    return dirent.isFile() && !dirent.name.startsWith('index') && hasAvailableExtension(dirent.name);
};
async function getExportables(cwd, excludeKeys) {
    const pathname = path__default.default.resolve(cwd, SRC);
    const dirents = await fs__default.default.readdir(pathname, {
        withFileTypes: true
    });
    const exportables = await Promise.all(dirents.map(async (dirent)=>await isExportable(dirent, pathname) && !excludeKeys.includes(dirent.name) ? dirent.name : undefined));
    return exportables.filter(nonNullable);
}
function mapWildcard(wildcardExports, exportables) {
    return exportables.map((exportable)=>{
        const isFile = exportable.includes('.');
        const filename = isFile ? filenameWithoutExtension(exportable) : exportable;
        return {
            [`./${filename}`]: Object.fromEntries(Object.entries(wildcardExports['./*']).map(([key, value])=>[
                    key,
                    value.replace(/\*/g, isFile ? filename : `${filename}/index`)
                ]))
        };
    });
}
async function resolveWildcardExports(exportsCondition, cwd) {
    if (!exportsCondition || typeof exportsCondition === 'string') return undefined;
    const hasWildcard = !!exportsCondition['./*'];
    if (hasWildcard) {
        console.warn(`The wildcard export "./*" is experimental and may change or be removed at any time.\n` + 'To open an issue, please visit https://github.com/huozhi/bunchee/issues' + '.\n');
        // './foo' -> ['foo']; './foo/bar' -> ['bar']
        // will contain '*' also but it's not a problem
        const excludeKeys = Object.keys(exportsCondition).map((key)=>key.split('/').pop());
        const exportables = await getExportables(cwd, excludeKeys);
        if (exportables.length > 0) {
            const wildcardExports = getWildcardExports(exportsCondition);
            const resolvedWildcardExports = mapWildcard(wildcardExports, exportables);
            const resolvedExports = Object.assign({}, exportsCondition, ...resolvedWildcardExports);
            delete resolvedExports['./*'];
            return resolvedExports;
        }
    }
    return undefined;
}

function assignDefault(options, name, defaultValue) {
    if (!(name in options) || options[name] == null) {
        options[name] = defaultValue;
    }
}
function hasMultiEntryExport(exportPaths) {
    const exportKeys = Object.keys(exportPaths).filter((key)=>key !== './package.json');
    return exportKeys.length > 0 && exportKeys.every((name)=>name.startsWith('.'));
}
async function bundle(entryPath, { cwd: _cwd, ...options } = {}) {
    var _tsConfig, _tsConfig1, _tsConfig2;
    const cwd = path.resolve(process.cwd(), _cwd || '');
    assignDefault(options, 'format', 'es');
    assignDefault(options, 'minify', false);
    assignDefault(options, 'target', 'es2015');
    const pkg = await getPackageMeta(cwd);
    const resolvedWildcardExports = await resolveWildcardExports(pkg.exports, cwd);
    const packageType = getPackageType(pkg);
    const exportPaths = getExportPaths(pkg, packageType, resolvedWildcardExports);
    const exportKeys = Object.keys(exportPaths).filter((key)=>key !== './package.json');
    // const exportPathsLength = Object.keys(exportPaths).length
    const isMultiEntries = hasMultiEntryExport(exportPaths) // exportPathsLength > 1
    ;
    const tsConfig = await resolveTsConfig(cwd);
    const hasTsConfig = Boolean((_tsConfig = tsConfig) == null ? void 0 : _tsConfig.tsConfigPath);
    const defaultTsOptions = {
        tsConfigPath: (_tsConfig1 = tsConfig) == null ? void 0 : _tsConfig1.tsConfigPath,
        tsCompilerOptions: ((_tsConfig2 = tsConfig) == null ? void 0 : _tsConfig2.tsCompilerOptions) || {}
    };
    // Handle single entry file
    if (!isMultiEntries) {
        // Use specified string file path if possible, then fallback to the default behavior entry picking logic
        // e.g. "exports": "./dist/index.js" -> use "./index.<ext>" as entry
        entryPath = entryPath || await getSourcePathFromExportPath(cwd, '.', 'default') || '';
    }
    if (entryPath) {
        let mainEntryPath;
        let typesEntryPath;
        // with -o option
        if (options.file) {
            mainEntryPath = options.file;
        } else if (exportKeys.length === 0) {
            mainEntryPath = path.resolve(cwd, 'dist/index.js');
        }
        if (mainEntryPath) {
            if (options.dts) {
                typesEntryPath = getTypeFilePath(mainEntryPath, undefined, cwd);
            }
            exportPaths['.'] = constructDefaultExportCondition({
                main: mainEntryPath,
                types: typesEntryPath
            }, packageType);
        }
    }
    const bundleOrWatch = (rollupConfig)=>{
        const { input, exportName } = rollupConfig;
        const exportPath = getExportPath(pkg, cwd, exportName);
        // Log original entry file relative path
        const source = typeof input.input === 'string' ? path.relative(cwd, input.input) : exportPath;
        const buildMetadata = {
            source
        };
        if (options.watch) {
            return Promise.resolve(runWatch(rollupConfig, buildMetadata));
        }
        return runBundle(rollupConfig);
    };
    const hasSpecifiedEntryFile = entryPath ? await fileExists(entryPath) && (await fs__default.default.stat(entryPath)).isFile() : false;
    if (!hasSpecifiedEntryFile && !isMultiEntries) {
        const err = new Error(`Entry file \`${entryPath}\` is not existed`);
        err.name = 'NOT_EXISTED';
        return Promise.reject(err);
    }
    let result;
    const buildConfigs = await buildEntryConfig(pkg, entryPath, exportPaths, options, cwd, defaultTsOptions, false);
    const assetsJobs = buildConfigs.map((rollupConfig)=>bundleOrWatch(rollupConfig));
    const typesJobs = hasTsConfig ? (await buildEntryConfig(pkg, entryPath, exportPaths, options, cwd, defaultTsOptions, true)).map((rollupConfig)=>bundleOrWatch(rollupConfig)) : [];
    result = await Promise.all(assetsJobs.concat(typesJobs));
    logSizeStats();
    return result;
}
function runWatch({ input, output }, metadata) {
    const watchOptions = [
        {
            ...input,
            output: output,
            watch: {
                exclude: [
                    'node_modules/**'
                ]
            }
        }
    ];
    const watcher = rollup.watch(watchOptions);
    watcher.on('event', (event)=>{
        switch(event.code){
            case 'ERROR':
                {
                    return onError(event.error);
                }
            case 'START':
                {
                    logger.log(`Start building ${metadata.source} ...`);
                    break;
                }
            case 'END':
                {
                    break;
                }
            default:
                return;
        }
    });
    return watcher;
}
function runBundle({ input, output }) {
    return rollup.rollup(input).then((bundle)=>{
        const writeJobs = output.map((options)=>bundle.write(options));
        return Promise.all(writeJobs);
    }, onError);
}
function onError(error) {
    if (!error) return;
    // logging source code in format
    if (error.frame) {
        process.stderr.write(error.frame + '\n');
    }
    // filter out the rollup plugin error information such as loc/frame/code...
    const err = new Error(error.message);
    err.stack = error.stack;
    throw err;
}

exports.bundle = bundle;
